#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2022 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External modules
import numpy as np
from scipy import ndimage
from hausdorff import hausdorff_distance

#-----------------------------------------------------#
#        Calculate : Simple Hausdorff Distance        #
#-----------------------------------------------------#
"""
References:
    A. A. Taha and A. Hanbury.
    "An Efficient Algorithm for Calculating the Exact Hausdorff Distance".
    IEEE Transactions on Pattern Analysis and Machine Intelligence,
    vol. 37, no. 11, pp. 2153-2163, 1 Nov. 2015, doi: 10.1109/TPAMI.2015.2408351.

Implementation: https://github.com/mavillan/py-hausdorff
"""
def calc_SimpleHausdorffDistance(truth, pred, c=1, **kwargs):
    # Compute simple Hausdorff Distance
    hd = hausdorff_distance(truth, pred, distance="euclidean")
    # Return Hausdorff Distance
    return np.float64(hd)

#-----------------------------------------------------#
#       Calculate : Average Hausdorff Distance        #
#-----------------------------------------------------#
"""
References:
    Adel Kermi, Issam Mahmoudi, Mohamed Tarek Khadir. 2019.
    Deep Convolutional Neural Networks Using U-Net for Automatic Brain Tumor Segmentation in Multimodal MRI Volumes.
    https://link.springer.com/chapter/10.1007/978-3-030-11726-9_4

Implementation: https://github.com/Issam28/Brain-tumor-segmentation
"""
def border_map(binary_img,neigh):
    """
    Creates the border for a 3D image
    """
    binary_map = np.asarray(binary_img, dtype=np.uint8)
    neigh = neigh
    west = ndimage.shift(binary_map, [-1, 0], order=0)
    east = ndimage.shift(binary_map, [1, 0], order=0)
    north = ndimage.shift(binary_map, [0, 1], order=0)
    south = ndimage.shift(binary_map, [0, -1], order=0)
    cumulative = west + east + north + south
    border = ((cumulative < 4) * binary_map) == 1
    return border

def border_distance(ref,seg):
    """
    This functions determines the map of distance from the borders of the
    segmentation and the reference and the border maps themselves
    """
    neigh=8
    border_ref = border_map(ref,neigh)
    border_seg = border_map(seg,neigh)
    oppose_ref = 1 - ref
    oppose_seg = 1 - seg
    # euclidean distance transform
    distance_ref = ndimage.distance_transform_edt(oppose_ref)
    distance_seg = ndimage.distance_transform_edt(oppose_seg)
    distance_border_seg = border_ref * distance_seg
    distance_border_ref = border_seg * distance_ref
    return distance_border_ref, distance_border_seg

def calc_AverageHausdorffDistance(truth, pred, c=1, **kwargs):
    """
    This functions calculates the average symmetric distance and the
    hausdorff distance between a segmentation and a reference image
    :return: hausdorff distance and average symmetric distance
    """
    # Obtain sets with associated class
    ref = np.equal(truth, c)
    seg = np.equal(pred, c)
    # Compute AHD
    ref_border_dist, seg_border_dist = border_distance(ref, seg)
    hausdorff_distance = np.max([np.max(ref_border_dist),
                                 np.max(seg_border_dist)])
    # Return AHD
    return hausdorff_distance



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2022 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External modules
import numpy as np
# Internal modules
from miseval.confusion_matrix import calc_ConfusionMatrix

#-----------------------------------------------------#
#              Calculate : Cohen's Kappa              #
#-----------------------------------------------------#
"""
References:
    Cohen J. A Coefficient of Agreement for Nominal Scales.
    Educ Psychol Meas [Internet]. 1960 Apr 2 [cited 2022 Jan 8];
    20(1):37–46. Available from:
    http://journals.sagepub.com/doi/10.1177/001316446002000104
"""
def calc_Kappa(truth, pred, c=1, dtype=np.float64, **kwargs):
    # Obtain confusion mat
    tp, tn, fp, fn = calc_ConfusionMatrix(truth, pred, c, dtype)
    # Compute kappa
    fa = tp + tn
    fc = ((tn+fn)*(tn+fp) + (fp+tp)*(fn+tp)) / (tp+tn+fp+fn)
    if ((tp+tn+fp+fn)-fc) != 0 : kappa = (fa-fc) / ((tp+tn+fp+fn)-fc)
    else : kappa = 0.0
    # Return kappa score
    return kappa



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2022 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2022 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External modules
import numpy as np
# Internal modules
from miseval.confusion_matrix import calc_ConfusionMatrix

#-----------------------------------------------------#
#              Calculate : DSC Enhanced               #
#-----------------------------------------------------#
"""
    Reference:  Dominik Müller, Adrian Pfleiderer & Frank Kramer. (2022).
                miseval: a metric library for Medical Image Segmentation EVALuation.
                https://github.com/frankkramer-lab/miseval

    Custom Dice Similarity Coefficient implementation which returns 1.0,
    if two empty masks are compared.
    This allow rewarding models which correctly predict empty masks.
"""
def calc_DSC_Enhanced(truth, pred, c=1, **kwargs):
    # Obtain sets with associated class
    gt = np.equal(truth, c)
    pd = np.equal(pred, c)
    # Calculate Dice
    if gt.sum() == 0 and pd.sum() == 0 : dice = 1.0
    elif (pd.sum() + gt.sum()) != 0:
        dice = 2*np.logical_and(pd, gt).sum() / (pd.sum() + gt.sum())
    else : dice = 0.0
    # Return computed Dice
    return dice

#-----------------------------------------------------#
#              Calculate : DSC via Sets               #
#-----------------------------------------------------#
def calc_DSC_Sets(truth, pred, c=1, **kwargs):
    # Obtain sets with associated class
    gt = np.equal(truth, c)
    pd = np.equal(pred, c)
    # Calculate Dice
    if (pd.sum() + gt.sum()) != 0:
        dice = 2*np.logical_and(pd, gt).sum() / (pd.sum() + gt.sum())
    else : dice = 0.0
    # Return computed Dice
    return dice

#-----------------------------------------------------#
#             Calculate : DSC via ConfMat             #
#-----------------------------------------------------#
def calc_DSC_CM(truth, pred, c=1, **kwargs):
    # Obtain confusion mat
    tp, tn, fp, fn = calc_ConfusionMatrix(truth, pred, c)
    # Calculate Dice
    if (2*tp + fp + fn) != 0 : dice = 2*tp / (2*tp + fp + fn)
    else : dice = 0.0
    # Return computed Dice
    return dice



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2022 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External modules
import numpy as np
# Internal modules
from miseval import metric_dict

#-----------------------------------------------------#
#             Core Hub Function: Evaluate             #
#-----------------------------------------------------#
""" The miseval eavluate function can be run with different metrics as backbone.
    You can pass the following options to the metric parameter:
    - String naming one of the metric labels, for example "DSC"
    - Directly passing a metric function, for example calc_DSC (from dice.py)
    - Passing a custom metric function

    List of metrics : See miseval/__init__.py under section "Access Functions to Metric Functions"

    The classes in a segmentation mask must be ongoing starting from 0 (integers from 0 to n_classes-1).

    A segmentation mask is allowed to have either no channel axis or just 1 (e.g. 512x512x1),
    which contains the annotation.

Arguments:
    truth (NumPy Matrix):            Ground Truth segmentation mask.
    pred (NumPy Matrix):             Prediction segmentation mask.
    metric (String or Function):     Metric function. Either a function directly or encoded as String from miseval or a custom function.
    multi_class (Boolean):           Boolean parameter, if segmentation is a binary or multi-class problem. By default False -> Binary mode.
    n_classes (Integer):             Number of classes. By default 2 -> Binary
    kwargs (arguments):              Additional arguments for passing down to metric functions.

Output:
    score (Float) or scores (List of Float)

    The multi_class parameter defines the output of this function.
    If n_classes > 2, multi_class is automatically True.
    If multi_class == False & n_classes == 2, only a single score (float) is returned.
    If multi_class == True, multiple scores as a list are returned (for each class one score).
"""
def evaluate(truth, pred, metric, multi_class=False, n_classes=2, **kwargs):
    # Initialize metric function
    if isinstance(metric, str):
        if metric in metric_dict : eval_metric = metric_dict[metric]
        elif metric.upper() in metric_dict:
            eval_metric = metric_dict[metric.upper()]
        else : raise KeyError("Provided metric string not in metric_dict!" + \
                              " : " + metric)
    elif callable(metric) : eval_metric = metric
    else : raise ValueError("Provided metric is neither a function nor a " + \
                            "string!" + " : " + str(metric))
    # Check some Exceptions
    if n_classes == 2 and len(np.unique(truth)) > 2:
        raise ValueError("Segmentation mask (truth) contains more than 2 classes!")
    if n_classes == 2 and len(np.unique(pred)) > 2:
        raise ValueError("Segmentation mask (pred) contains more than 2 classes!")
    # Run binary mode       -> Compute score only for main class
    if not multi_class and n_classes == 2:
        score = eval_metric(truth, pred, c=1, **kwargs)
        return score
    # Run multi-class mode  -> Compute score for each class
    else:
        score_list = np.zeros((n_classes,))
        for c in range(n_classes):
            score = eval_metric(truth, pred, c=c, **kwargs)
            score_list[c] = score
        return score_list



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2022 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External modules
import numpy as np
import unittest
# Internal modules
from miseval import *

#-----------------------------------------------------#
#             Unittest: Boundary Distance             #
#-----------------------------------------------------#
class TEST_BoundaryDistance(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        # Create ground truth
        np.random.seed(1)
        self.gt_bi = np.random.randint(2, size=(16,16))
        self.gt_mc = np.random.randint(5, size=(16,16))
        # Create prediction mask
        np.random.seed(2)
        self.pd_bi = np.random.randint(2, size=(16,16))
        self.pd_mc = np.random.randint(5, size=(16,16))
        # Create prediction probability
        self.prob_bi = np.random.rand(16,16,2)
        self.prob_mc = np.random.rand(16,16,5)

    #-------------------------------------------------#
    #          Calculate : Boundary Distance          #
    #-------------------------------------------------#
    def test_calc_BoundaryDistance_General(self):
        # Check binary score
        score_bi = calc_Boundary_Distance(self.gt_bi, self.pd_bi, c=1)
        self.assertTrue(isinstance(score_bi, np.float64))
        # Check multi-class score
        for i in range(5):
            score_mc = calc_Boundary_Distance(self.gt_mc, self.pd_mc, c=i)
            self.assertTrue(isinstance(score_mc, np.float64))
        # Check existance in metric_dict
        self.assertTrue("BoundaryDistance" in metric_dict)
        self.assertTrue(callable(metric_dict["BoundaryDistance"]))
        self.assertTrue("Distance" in metric_dict)
        self.assertTrue(callable(metric_dict["Distance"]))
        self.assertTrue("BD" in metric_dict)
        self.assertTrue(callable(metric_dict["BD"]))

    #-------------------------------------------------#
    #      Boundary Distance : Distance Function      #
    #-------------------------------------------------#
    def test_calc_BoundaryDistance_DistanceFunction(self):
        # Define list of pooling function
        df_list = ["bhattacharyya", "bhattacharyya_coefficient", "canberra",
                   "chebyshev", "chi_square", "cosine", "euclidean", "hamming",
                   "jensen_shannon", "kullback_leibler", "mae", "manhattan",
                   "minkowsky", "mse", "pearson", "squared_variation"]
        # Test boundary distance with each pooling function
        for df in df_list:
            score_bi = calc_Boundary_Distance(self.gt_bi, self.pd_bi, c=1,
                                              distance=df)
            self.assertTrue(isinstance(score_bi, np.float64))

    #-------------------------------------------------#
    #       Boundary Distance : Pooling Function      #
    #-------------------------------------------------#
    def test_calc_BoundaryDistance_PoolingFunction(self):
        # Define list of pooling function
        pf_list = ["sum", "mean", "amin", "amax"]
        # Test boundary distance with each pooling function
        for pf in pf_list:
            score_bi = calc_Boundary_Distance(self.gt_bi, self.pd_bi, c=1,
                                              pooling=pf)
            self.assertTrue(isinstance(score_bi, np.float64))



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2022 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External modules
import numpy as np
from sklearn.metrics import balanced_accuracy_score, adjusted_rand_score
import warnings
# Internal modules
from miseval.confusion_matrix import calc_ConfusionMatrix

#-----------------------------------------------------#
#            Calculate : Accuracy via Sets            #
#-----------------------------------------------------#
def calc_Accuracy_Sets(truth, pred, c=1, **kwargs):
    # Obtain sets with associated class
    gt = np.equal(truth, c)
    pd = np.equal(pred, c)
    not_gt = np.logical_not(gt)
    not_pd = np.logical_not(pd)
    # Calculate Accuracy
    acc = (np.logical_and(pd, gt).sum() + \
           np.logical_and(not_pd, not_gt).sum()) / gt.size
    # Return computed Accuracy
    return acc

#-----------------------------------------------------#
#           Calculate : Accuracy via ConfMat          #
#-----------------------------------------------------#
def calc_Accuracy_CM(truth, pred, c=1, **kwargs):
    # Obtain confusion mat
    tp, tn, fp, fn = calc_ConfusionMatrix(truth, pred, c)
    # Calculate Accuracy
    acc = (tp + tn) / (tp + tn + fp + fn)
    # Return computed Accuracy
    return acc

#-----------------------------------------------------#
#            Calculate : Balanced Accuracy            #
#-----------------------------------------------------#
"""
Formula:
    BACC = (Sensitivity + Specificity) / 2

References:
[1] Brodersen, K.H.; Ong, C.S.; Stephan, K.E.; Buhmann, J.M. (2010).
    The balanced accuracy and its posterior distribution.
    Proceedings of the 20th International Conference on Pattern Recognition, 3121-24.

[2] John. D. Kelleher, Brian Mac Namee, Aoife D’Arcy, (2015).
    Fundamentals of Machine Learning for Predictive Data Analytics: Algorithms, Worked Examples, and Case Studies.
    https://mitpress.mit.edu/books/fundamentals-machine-learning-predictive-data-analytics
"""
def calc_BalancedAccuracy(truth, pred, c=1, **kwargs):
    # Obtain sets with associated class
    gt = np.equal(truth, c).flatten()
    pd = np.equal(pred, c).flatten()
    # Compute BACC via scikit-learn
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        bacc = balanced_accuracy_score(gt, pd)
    # Return BACC score
    return np.float64(bacc)

#-----------------------------------------------------#
#           Calculate : Adjusted Rand Index           #
#-----------------------------------------------------#
"""
Formula:
    ARI = (RI - Expected_RI) / (max(RI) - Expected_RI)

References:
[1] L. Hubert and P. Arabie, Comparing Partitions, Journal of Classification 1985
    https://link.springer.com/article/10.1007%2FBF01908075


[2] D. Steinley, Properties of the Hubert-Arabie adjusted Rand index,
    Psychological Methods 2004

[3] https://en.wikipedia.org/wiki/Rand_index#Adjusted_Rand_index
"""
def calc_AdjustedRandIndex(truth, pred, c=1, **kwargs):
    # Obtain sets with associated class
    gt = np.equal(truth, c).flatten()
    pd = np.equal(pred, c).flatten()
    # Compute ARI via scikit-learn
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        ari = adjusted_rand_score(gt, pd)
    # Return ARI score
    return np.float64(ari)



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2022 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External modules
import numpy as np

#-----------------------------------------------------#
#                Calculate : Hinge loss               #
#-----------------------------------------------------#
""" Compute Hinge loss between truth and prediction probabilities.

    In machine learning, the hinge loss is a loss function used for training classifiers.
    The hinge loss is used for "maximum-margin" classification.

Pooling (how to combine computed Hinge losses to a single value):
    Distance Sum                        sum
    Distance Averaging                  mean
    Minimum Distance                    amin
    Maximum Distance                    amax
"""
def calc_Hinge(truth, pred_prob, c=1, pooling="mean", provided_prob=True,
               **kwargs):
    # Obtain binary classification
    if provided_prob : prob = np.take(pred_prob, c, axis=-1)
    else : prob = np.equal(pred_prob, c)
    gt = np.equal(truth, c).astype(int)
    # Convert ground truth 0/1 format to -1/+1 format
    gt = np.where(gt==0, -1, gt)
    # Compute Hinge
    hinge_total = np.maximum(1 - gt * prob, 0)
    # Apply pooling function across all pixel classifications
    hinge = getattr(np, pooling)(hinge_total)
    # Return Hinge
    return hinge



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2022 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External modules
import numpy as np
import math
# Internal modules
from miseval.confusion_matrix import calc_ConfusionMatrix

#-----------------------------------------------------#
#     Calculate : Matthews Correlation Coefficient    #
#-----------------------------------------------------#
"""
References:
    Chicco, D., Jurman, G. The advantages of the Matthews correlation coefficient
    (MCC) over F1 score and accuracy in binary classification evaluation.
    BMC Genomics 21, 6 (2020). https://doi.org/10.1186/s12864-019-6413-7
"""
def calc_MCC(truth, pred, c=1, dtype=np.float64, **kwargs):
    # Obtain confusion mat
    tp, tn, fp, fn = calc_ConfusionMatrix(truth, pred, c, dtype)
    # Verify if we need an approximation to zero (prove in reference)
    if (tp >= 1 and fp >= 1 and tn == 0 and fn == 0) or \
       (tp == 0 and fp >= 1 and tn >= 1 and fn == 0) or \
       (tp == 0 and fp == 0 and tn >= 1 and fn >= 1) or \
       (tp >= 1 and fp == 0 and tn == 0 and fn >= 1):
       return 0.0
    # else compute mcc
    top = tp*tn - fp*fn
    bot_raw = (tp+fp) * (tp+fn) * (tn+fp) * (tn+fn)
    bot = math.sqrt(bot_raw)
    if bot != 0 : mcc = top / bot
    else : mcc = 0.0
    # Return mcc score
    return mcc

#-----------------------------------------------------#
#             Calculate : Normalized MCC              #
#-----------------------------------------------------#
"""
References:
    Chicco, D., Jurman, G. The advantages of the Matthews correlation coefficient
    (MCC) over F1 score and accuracy in binary classification evaluation.
    BMC Genomics 21, 6 (2020). https://doi.org/10.1186/s12864-019-6413-7
"""
def calc_MCC_Normalized(truth, pred, c=1, **kwargs):
    # Compute mcc
    mcc = calc_MCC(truth, pred, c)
    # Normalize it
    nmcc = (mcc + 1) / 2
    # Return normalized mcc score
    return nmcc

#-----------------------------------------------------#
#               Calculate : Absolute MCC              #
#-----------------------------------------------------#
"""
References:
    Chicco, D., Jurman, G. The advantages of the Matthews correlation coefficient
    (MCC) over F1 score and accuracy in binary classification evaluation.
    BMC Genomics 21, 6 (2020). https://doi.org/10.1186/s12864-019-6413-7
"""
def calc_MCC_Absolute(truth, pred, c=1, **kwargs):
    # Compute mcc
    mcc = calc_MCC(truth, pred, c)
    # Absolute it
    amcc = abs(mcc)
    # Return normalized mcc score
    return amcc



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2022 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External modules
import numpy as np
import unittest
# Internal modules
from miseval import *

#-----------------------------------------------------#
#        Unittest: Dice Similarity Coefficient        #
#-----------------------------------------------------#
class TEST_DiceSimilarityCoefficient(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        # Create ground truth
        np.random.seed(1)
        self.gt_bi = np.random.randint(2, size=(32,32))
        self.gt_mc = np.random.randint(5, size=(32,32))
        # Create prediction
        np.random.seed(2)
        self.pd_bi = np.random.randint(2, size=(32,32))
        self.pd_mc = np.random.randint(5, size=(32,32))

    #-------------------------------------------------#
    #            Calculate : DSC Variant #1           #
    #-------------------------------------------------#
    def test_calc_DSC_Sets(self):
        # Check binary score
        score_bi = calc_DSC_Sets(self.gt_bi, self.pd_bi, c=1)
        self.assertTrue(isinstance(score_bi, np.float64))
        self.assertTrue(score_bi == calc_DSC(self.gt_bi, self.pd_bi, c=1))
        # Check multi-class score
        for i in range(5):
            score_mc = calc_DSC_Sets(self.gt_mc, self.pd_mc, c=i)
            self.assertTrue(isinstance(score_mc, np.float64))
            self.assertTrue(score_mc == calc_DSC(self.gt_mc, self.pd_mc, c=i))

    #-------------------------------------------------#
    #            Calculate : DSC Variant #2           #
    #-------------------------------------------------#
    def test_calc_DSC_CM(self):
        # Check binary score
        score_bi = calc_DSC_CM(self.gt_bi, self.pd_bi, c=1)
        self.assertTrue(isinstance(score_bi, np.float64))
        self.assertTrue(score_bi == calc_DSC(self.gt_bi, self.pd_bi, c=1))
        # Check multi-class score
        for i in range(5):
            score_mc = calc_DSC_CM(self.gt_mc, self.pd_mc, c=i)
            self.assertTrue(isinstance(score_mc, np.float64))
            self.assertTrue(score_mc == calc_DSC(self.gt_mc, self.pd_mc, c=i))

    #-------------------------------------------------#
    #            Calculate : DSC Variant #1           #
    #-------------------------------------------------#
    def test_calc_DSC_Enhanced(self):
        # Check binary score
        score_bi = calc_DSC_Enhanced(self.gt_bi, self.pd_bi, c=1)
        self.assertTrue(isinstance(score_bi, np.float64))
        # Check multi-class score
        for i in range(5):
            score_mc = calc_DSC_Enhanced(self.gt_mc, self.pd_mc, c=i)
            self.assertTrue(isinstance(score_mc, np.float64))
        # Check existance in metric_dict
        self.assertTrue("DSC" in metric_dict)
        self.assertTrue(callable(metric_dict["DSC"]))
        self.assertTrue("Dice" in metric_dict)
        self.assertTrue(callable(metric_dict["Dice"]))
        self.assertTrue("DiceSimilarityCoefficient" in metric_dict)
        self.assertTrue(callable(metric_dict["DiceSimilarityCoefficient"]))



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2022 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
import numpy as np

#-----------------------------------------------------#
#            Calculate : Confusion Matrix             #
#-----------------------------------------------------#
def calc_ConfusionMatrix(truth, pred, c=1, dtype=np.int64, **kwargs):
    # Obtain predicted and actual condition
    gt = np.equal(truth, c)
    pd = np.equal(pred, c)
    not_gt = np.logical_not(gt)
    not_pd = np.logical_not(pd)
    # Compute Confusion Matrix
    tp = np.logical_and(pd, gt).sum()
    tn = np.logical_and(not_pd, not_gt).sum()
    fp = np.logical_and(pd, not_gt).sum()
    fn = np.logical_and(not_pd, gt).sum()
    # Convert to desired numpy type to avoid overflow
    tp = tp.astype(dtype)
    tn = tn.astype(dtype)
    fp = fp.astype(dtype)
    fn = fn.astype(dtype)
    # Return Confusion Matrix
    return tp, tn, fp, fn

#-----------------------------------------------------#
#              Calculate : True Positive              #
#-----------------------------------------------------#
def calc_TruePositive(truth, pred, c=1, **kwargs):
    # Obtain predicted and actual condition
    gt = np.equal(truth, c)
    pd = np.equal(pred, c)
    not_gt = np.logical_not(gt)
    not_pd = np.logical_not(pd)
    # Compute true positive
    tp = np.logical_and(pd, gt).sum()
    # Return true positive
    return tp

#-----------------------------------------------------#
#              Calculate : True Negative              #
#-----------------------------------------------------#
def calc_TrueNegative(truth, pred, c=1, **kwargs):
    # Obtain predicted and actual condition
    gt = np.equal(truth, c)
    pd = np.equal(pred, c)
    not_gt = np.logical_not(gt)
    not_pd = np.logical_not(pd)
    # Compute true negative
    tn = np.logical_and(not_pd, not_gt).sum()
    # Return true negative
    return tn

#-----------------------------------------------------#
#              Calculate : False Positive             #
#-----------------------------------------------------#
def calc_FalsePositive(truth, pred, c=1, **kwargs):
    # Obtain predicted and actual condition
    gt = np.equal(truth, c)
    pd = np.equal(pred, c)
    not_gt = np.logical_not(gt)
    not_pd = np.logical_not(pd)
    # Compute false positive
    fp = np.logical_and(pd, not_gt).sum()
    # Return false positive
    return fp

#-----------------------------------------------------#
#              Calculate : False Negative             #
#-----------------------------------------------------#
def calc_FalseNegative(truth, pred, c=1, **kwargs):
    # Obtain predicted and actual condition
    gt = np.equal(truth, c)
    pd = np.equal(pred, c)
    not_gt = np.logical_not(gt)
    not_pd = np.logical_not(pd)
    # Compute false negative
    fn = np.logical_and(not_pd, gt).sum()
    # Return false negative
    return fn



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2022 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External modules
import numpy as np
from skimage.measure import find_contours
import dictances

#-----------------------------------------------------#
#            Calculate : Boundary Distance            #
#-----------------------------------------------------#
"""
Computes distance of segmentation boundaries between ground truth and prediction.

List of available distances:
    Bhattacharyya distance 	            bhattacharyya
    Bhattacharyya coefficient           bhattacharyya_coefficient
    Canberra distance 	                canberra
    Chebyshev distance 	                chebyshev
    Chi Square distance 	            chi_square
    Cosine Distance 	                cosine
    Euclidean distance 	                euclidean
    Hamming distance 	                hamming
    Jensen-Shannon divergence 	        jensen_shannon
    Kullback-Leibler divergence         kullback_leibler
    Mean absolute error 	            mae
    Taxicab geometry 	                manhattan, cityblock, total_variation
    Minkowski distance 	                minkowsky
    Mean squared error 	                mse
    Pearson's distance 	                pearson
    Squared deviations from the mean 	squared_variation

Distance Pooling (how to combine computed distances to a single value):
    Distance Sum                        sum
    Distance Averaging                  mean
    Minimum Distance                    amin
    Maximum Distance                    amax

Distance Implementation:
    Credits to Luca Cappelletti
    https://github.com/LucaCappelletti94/dictances

"""
def calc_Boundary_Distance(truth, pred, c=1, distance="euclidean",
                           pooling="mean", **kwargs):
    # Obtain sets with associated class
    gt = np.equal(truth, c)
    pd = np.equal(pred, c)
    # Initialize result list
    res_dist = []
    try:
        # Compute boundary map for gt & pd
        gt_bm = np.concatenate(find_contours(gt))
        pd_bm = np.concatenate(find_contours(pd))
        # Compute distance between each coord
        for i in range(gt_bm.shape[0]):
            for j in range(pd_bm.shape[0]):
                # Create dictionary as interface for dictances package
                a = {"x": gt_bm[i][0],
                     "y": gt_bm[i][1]}
                b = {"x": pd_bm[j][0],
                     "y": pd_bm[j][1]}
                # Compute distance via dictances
                dist = getattr(dictances, distance)(a, b)
                # Append to result list
                res_dist.append(dist)
    except Exception as e:
        print("Distance", distance, "does NOT support all edge cases:", e)
    # Apply pooling function
    res = getattr(np, pooling)(res_dist)
    # Return AUC
    return res



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2022 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External modules
import numpy as np
import unittest
# Internal modules
from miseval import evaluate, metric_dict

#-----------------------------------------------------#
#          Unittest: Edge Cases & Exceptions          #
#-----------------------------------------------------#
class TEST_EdgeCases(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        # Create identical ground truth & prediction
        np.random.seed(1)
        self.gt_identical = np.random.randint(2, size=(32,32))
        self.pd_identical = np.random.randint(2, size=(32,32))
        # Create normal ground truth & prediction
        np.random.seed(8)
        self.gt_normal = np.random.randint(2, size=(32,32))
        np.random.seed(16)
        self.pd_normal = np.random.randint(2, size=(32,32))
        # Create empty ground truth & prediction
        self.gt_empty = np.zeros((32,32))
        self.pd_empty = np.zeros((32,32))
        # Create full ground truth & prediction
        self.gt_full = np.full((32,32), 1)
        self.pd_full = np.full((32,32), 1)

    #-------------------------------------------------#
    #              EdgeCase : Variant #1              #
    #-------------------------------------------------#
    def test_EdgeCase_identical(self):
        for metric in metric_dict:
            scores = evaluate(self.gt_identical, self.pd_identical, metric,
                             multi_class=True, n_classes=2,
                             provided_prob=False)
            self.assertTrue(isinstance(scores[0], np.int64) or \
                            isinstance(scores[0], np.float64))
            self.assertTrue(isinstance(scores[1], np.int64) or \
                            isinstance(scores[1], np.float64))

    #-------------------------------------------------#
    #              EdgeCase : Variant #2              #
    #-------------------------------------------------#
    def test_EdgeCase_normal_vs_all(self):
        for metric in metric_dict:
            gt = self.gt_normal
            for pd in [self.pd_normal, self.pd_empty, self.pd_full]:
                scores = evaluate(gt, pd, metric,
                                  multi_class=True, n_classes=2,
                                  provided_prob=False)
                self.assertTrue(isinstance(scores[0], np.int64) or \
                                isinstance(scores[0], np.float64))
                self.assertTrue(isinstance(scores[1], np.int64) or \
                                isinstance(scores[1], np.float64))

    #-------------------------------------------------#
    #              EdgeCase : Variant #3              #
    #-------------------------------------------------#
    def test_EdgeCase_empty_vs_all(self):
        for metric in metric_dict:
            gt = self.gt_empty
            for pd in [self.pd_normal, self.pd_empty, self.pd_full]:
                scores = evaluate(gt, pd, metric,
                                  multi_class=True, n_classes=2,
                                  provided_prob=False)
                self.assertTrue(isinstance(scores[0], np.int64) or \
                                isinstance(scores[0], np.float64))
                self.assertTrue(isinstance(scores[1], np.int64) or \
                                isinstance(scores[1], np.float64))

    #-------------------------------------------------#
    #              EdgeCase : Variant #4              #
    #-------------------------------------------------#
    def test_EdgeCase_full_vs_all(self):
        for metric in metric_dict:
            gt = self.gt_full
            for pd in [self.pd_normal, self.pd_empty, self.pd_full]:
                scores = evaluate(gt, pd, metric,
                                  multi_class=True, n_classes=2,
                                  provided_prob=False)
                self.assertTrue(isinstance(scores[0], np.int64) or \
                                isinstance(scores[0], np.float64))
                self.assertTrue(isinstance(scores[1], np.int64) or \
                                isinstance(scores[1], np.float64))



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
   name='miseval',
   version='1.2.0',
   description='A Metric Library for Medical Image Segmentation Evaluation',
   url='https://github.com/frankkramer-lab/miseval',
   author='Dominik Müller',
   author_email='dominik.mueller@informatik.uni-augsburg.de',
   license='GPLv3',
   long_description=long_description,
   long_description_content_type="text/markdown",
   packages=find_packages(),
   python_requires='>=3.8',
   install_requires=['numpy<1.22,>=1.18',
                     'scikit-learn>=1.0.2',
                     'scikit-image>=0.19.1',
                     'scipy>=1.7.3',
                     'hausdorff>=0.2.6',
                     'numba>=0.54.0',
                     'dictances>=1.5.3'],
   classifiers=["Programming Language :: Python :: 3",
                "Programming Language :: Python :: 3.8",
                "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
                "Operating System :: OS Independent",

                "Intended Audience :: Healthcare Industry",
                "Intended Audience :: Science/Research",

                "Topic :: Scientific/Engineering :: Artificial Intelligence",
                "Topic :: Scientific/Engineering :: Image Recognition",
                "Topic :: Scientific/Engineering :: Medical Science Apps.",
                "Topic :: Scientific/Engineering :: Mathematics"]
)



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2022 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External modules
import numpy as np
from sklearn.metrics import roc_auc_score
# Internal modules
from miseval.confusion_matrix import calc_ConfusionMatrix

#-----------------------------------------------------#
#            Calculate : AUC via trapezoid            #
#-----------------------------------------------------#
"""
Formula:
    AUC = 1 - 1/2 * (FP/(FP+TN) + FN/(FN+TP))

References:
    Powers DMW. Evaluation: from precision, recall and F-measure to ROC, informedness, markedness and correlation.
    2020 Oct 10 [cited 2022 Jan 8]; Available from: http://arxiv.org/abs/2010.16061
"""
def calc_AUC_trapezoid(truth, pred, c=1, **kwargs):
    # Obtain confusion mat
    tp, tn, fp, fn = calc_ConfusionMatrix(truth, pred, c)
    # Compute AUC
    if (fp+tn) != 0 : x = fp/(fp+tn)
    else : x = 0.0
    if (fn+tp) != 0 : y = fn/(fn+tp)
    else : y = 0.0
    auc = 1 - (1/2)*(x + y)
    # Return AUC
    return auc

#-----------------------------------------------------#
#           Calculate : AUC via probability           #
#-----------------------------------------------------#
def calc_AUC_probability(truth, pred_prob, c=1, rounding_precision=5, **kwargs):
    # Round probability to reduce unnecessary thresholds
    prob = np.round(pred_prob[:,:,c], rounding_precision)
    # Obtain ground truth set with associated class
    gt = np.equal(truth, c).astype(int)
    auc = roc_auc_score(gt.flatten(), prob.flatten())
    # Return AUC
    return auc



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2022 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External modules
import numpy as np

#-----------------------------------------------------#
#              Calculate : Cross-Entropy              #
#-----------------------------------------------------#
""" Compute cross-entropy between truth and prediction probabilities.

    In information theory, the cross-entropy between two probability distributions p and q
    over the same underlying set of events measures the average number of bits needed to
    identify an event drawn from the set if a coding scheme used for the set is optimized
    for an estimated probability distribution q, rather than the true distribution p.

    Source: https://en.wikipedia.org/wiki/Cross_entropy

Pooling (how to combine computed cross-entropy to a single value):
    Distance Sum                        sum
    Distance Averaging                  mean
    Minimum Distance                    amin
    Maximum Distance                    amax
"""
def calc_CrossEntropy(truth, pred_prob, c=1, pooling="mean", provided_prob=True,
                      **kwargs):
    # Obtain binary classification
    if provided_prob : prob = np.take(pred_prob, c, axis=-1)
    else : prob = np.equal(pred_prob, c)
    gt = np.equal(truth, c).astype(int)
    # Add epsilon to probability to avoid zero divisions for log()
    prob = prob + np.finfo(np.float32).eps
    # Compute cross-entropy
    cross_entropy = - gt * np.log(prob)
    # Apply pooling function across all pixel classifications
    res = getattr(np, pooling)(cross_entropy)
    # Return Cross-Entropy
    return res
